@@ -55,6 +55,7 @@ struct _BluetoothChooserPrivate {
 
 	/* Widgets/UI bits that can be shown or hidden */
 	GtkCellRenderer *bonded_cell;
+	GtkCellRenderer *connected_cell;
 	GtkWidget *treeview;
 	GtkWidget *search_button;
 	GtkWidget *device_type_label, *device_type;
@@ -63,9 +64,11 @@ struct _BluetoothChooserPrivate {
 
 	/* Current filter */
 	int device_type_filter;
+	GtkTreeModel *device_type_filter_model;
 	int device_category_filter;
 
 	guint show_paired : 1;
+	guint show_connected : 1;
 	guint show_search : 1;
 	guint show_device_type : 1;
 	guint show_device_category : 1;
@@ -74,6 +77,12 @@ struct _BluetoothChooserPrivate {
 
 G_DEFINE_TYPE(BluetoothChooser, bluetooth_chooser, GTK_TYPE_VBOX)
 
+enum {
+	DEVICE_TYPE_FILTER_COL_NAME = 0,
+	DEVICE_TYPE_FILTER_COL_MASK,
+	DEVICE_TYPE_FILTER_NUM_COLS
+};
+
 static const char *
 bluetooth_device_category_to_string (int type)
 {
@@ -86,20 +95,13 @@ bluetooth_device_category_to_string (int type)
 		return N_("Trusted");
 	case BLUETOOTH_CATEGORY_NOT_PAIRED_OR_TRUSTED:
 		return N_("Not paired or trusted");
+	case BLUETOOTH_CATEGORY_PAIRED_OR_TRUSTED:
+		return N_("Paired or trusted");
 	default:
 		return N_("Unknown");
 	}
 }
 
-static int
-int_log2(int v)
-{
-	int rv = 0;
-	while (v >>= 1)
-		rv++;
-	return rv;
-}
-
 static void
 bonded_to_icon (GtkTreeViewColumn *column, GtkCellRenderer *cell,
 	      GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
@@ -108,10 +110,18 @@ bonded_to_icon (GtkTreeViewColumn *column, GtkCellRenderer *cell,
 
 	gtk_tree_model_get (model, iter, BLUETOOTH_COLUMN_PAIRED, &bonded, -1);
 
-	if (bonded == FALSE)
-		g_object_set (cell, "stock-id", NULL, NULL);
-	else
-		g_object_set (cell, "stock-id", GTK_STOCK_DIALOG_AUTHENTICATION, NULL);
+	g_object_set (cell, "icon-name", bonded ? "bluetooth-paired" : NULL, NULL);
+}
+
+static void
+connected_to_icon (GtkTreeViewColumn *column, GtkCellRenderer *cell,
+	      GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
+{
+	gboolean connected;
+
+	gtk_tree_model_get (model, iter, BLUETOOTH_COLUMN_CONNECTED, &connected, -1);
+
+	g_object_set (cell, "icon-name", connected ? GTK_STOCK_CONNECT : NULL, NULL);
 }
 
 static void
@@ -121,10 +131,8 @@ type_to_text (GtkTreeViewColumn *column, GtkCellRenderer *cell,
 	guint type;
 
 	gtk_tree_model_get (model, iter, BLUETOOTH_COLUMN_TYPE, &type, -1);
-	if (type == 0)
-		g_object_set (cell, "text", _("Unknown"), NULL);
-	else
-		g_object_set (cell, "text", bluetooth_type_to_string (type), NULL);
+
+	g_object_set (cell, "text", (type == 0) ? _("Unknown") : bluetooth_type_to_string (type), NULL);
 }
 
 /**
@@ -224,6 +232,27 @@ bluetooth_chooser_get_selected_device_type (BluetoothChooser *self)
 	return type;
 }
 
+/**
+ * bluetooth_chooser_get_selected_device_is_connected:
+ * @self: a #BluetoothChooser widget
+ *
+ * Return value: whether the selected device is conncted to this computer, will be %FALSE if no devices are selected
+ */
+gboolean
+bluetooth_chooser_get_selected_device_is_connected (BluetoothChooser *self)
+{
+	BluetoothChooserPrivate *priv = BLUETOOTH_CHOOSER_GET_PRIVATE(self);
+	GtkTreeIter iter;
+	gboolean selected, connected;
+
+	selected = gtk_tree_selection_get_selected (priv->selection, NULL, &iter);
+	if (selected == FALSE)
+		return 0;
+
+	gtk_tree_model_get (priv->filter, &iter, BLUETOOTH_COLUMN_CONNECTED, &connected, -1);
+	return connected;
+}
+
 /**
  * bluetooth_chooser_set_title:
  * @self: a BluetoothChooser widget
@@ -233,15 +262,17 @@ void
 bluetooth_chooser_set_title (BluetoothChooser  *self, const char *title)
 {
 	BluetoothChooserPrivate *priv = BLUETOOTH_CHOOSER_GET_PRIVATE(self);
-	char *str;
 
 	if (title == NULL) {
-		str = g_strdup_printf ("<b>%s</b>", _("Recognized Bluetooth Devices"));
+		gtk_widget_hide (priv->label);
 	} else {
+		char *str;
+
 		str = g_strdup_printf ("<b>%s</b>", title);
+		gtk_label_set_markup (GTK_LABEL(priv->label), str);
+		g_free (str);
+		gtk_widget_show (priv->label);
 	}
-	gtk_label_set_markup (GTK_LABEL(priv->label), str);
-	g_free (str);
 }
 
 static void
@@ -264,6 +295,7 @@ device_model_row_changed (GtkTreeModel *model,
 	/* Maybe it's the name that changed */
 	if (name != NULL)
 		g_object_notify (G_OBJECT (self), "device-selected-name");
+	g_object_notify (G_OBJECT (self), "device-selected-is-connected");
 
 	g_free (name);
 }
@@ -321,6 +353,8 @@ filter_category_func (GtkTreeModel *model, GtkTreeIter *iter, BluetoothChooserPr
 		return trusted;
 	if (priv->device_category_filter == BLUETOOTH_CATEGORY_NOT_PAIRED_OR_TRUSTED)
 		return (!bonded && !trusted);
+	if (priv->device_category_filter == BLUETOOTH_CATEGORY_PAIRED_OR_TRUSTED)
+		return (bonded || trusted);
 
 	g_assert_not_reached ();
 
@@ -341,12 +375,63 @@ filter_type_changed_cb (GtkComboBox *widget, gpointer data)
 {
 	BluetoothChooser *self = BLUETOOTH_CHOOSER (data);
 	BluetoothChooserPrivate *priv = BLUETOOTH_CHOOSER_GET_PRIVATE(self);
+	GtkTreeIter iter;
+
+	if (gtk_combo_box_get_active_iter (widget, &iter) == FALSE)
+		return;
+
+	gtk_tree_model_get (GTK_TREE_MODEL (priv->device_type_filter_model), &iter,
+			    DEVICE_TYPE_FILTER_COL_MASK, &(priv->device_type_filter),
+			    -1);
 
-	priv->device_type_filter = 1 << gtk_combo_box_get_active (GTK_COMBO_BOX(priv->device_type));
 	if (priv->filter)
 		gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER (priv->filter));
 	g_object_notify (G_OBJECT(self), "device-type-filter");
 }
+
+static void
+set_combobox_from_filter (BluetoothChooser *self)
+{
+	BluetoothChooserPrivate *priv = BLUETOOTH_CHOOSER_GET_PRIVATE(self);
+	GtkTreeIter iter;
+	gboolean cont;
+
+	/* Look for an exact matching filter first */
+	cont = gtk_tree_model_get_iter_first (GTK_TREE_MODEL (priv->device_type_filter_model),
+					      &iter);
+	while (cont != FALSE) {
+		int mask;
+
+		gtk_tree_model_get (GTK_TREE_MODEL (priv->device_type_filter_model), &iter,
+				    DEVICE_TYPE_FILTER_COL_MASK, &mask, -1);
+		if (mask == priv->device_type_filter) {
+			gtk_combo_box_set_active_iter (GTK_COMBO_BOX(priv->device_type), &iter);
+			return;
+		}
+		cont = gtk_tree_model_iter_next (GTK_TREE_MODEL (priv->device_type_filter_model), &iter);
+	}
+
+	/* Then a fuzzy match */
+	cont = gtk_tree_model_get_iter_first (GTK_TREE_MODEL (priv->device_type_filter_model),
+					      &iter);
+	while (cont != FALSE) {
+		int mask;
+
+		gtk_tree_model_get (GTK_TREE_MODEL (priv->device_type_filter_model), &iter,
+				    DEVICE_TYPE_FILTER_COL_MASK, &mask,
+				    -1);
+		if (mask & priv->device_type_filter) {
+			gtk_combo_box_set_active_iter (GTK_COMBO_BOX(priv->device_type), &iter);
+			return;
+		}
+		cont = gtk_tree_model_iter_next (GTK_TREE_MODEL (priv->device_type_filter_model), &iter);
+	}
+
+	/* Then just set the any then */
+	gtk_tree_model_get_iter_first (GTK_TREE_MODEL (priv->device_type_filter_model), &iter);
+	gtk_combo_box_set_active_iter (GTK_COMBO_BOX(priv->device_type), &iter);
+}
+
 static void
 filter_category_changed_cb (GtkComboBox *widget, gpointer data)
 {
@@ -417,6 +502,8 @@ static void default_adapter_changed (GObject    *gobject,
 		gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER (priv->filter),
 							filter_func, self, NULL);
 		gtk_tree_view_set_model (GTK_TREE_VIEW(priv->treeview), priv->filter);
+		g_signal_connect (priv->filter, "row-changed",
+				  G_CALLBACK (device_model_row_changed), self);
 		g_object_unref (priv->filter);
 		gtk_widget_set_sensitive (GTK_WIDGET (priv->treeview), TRUE);
 		gtk_widget_set_sensitive (GTK_WIDGET(priv->search_button), TRUE);
@@ -427,9 +514,6 @@ static void default_adapter_changed (GObject    *gobject,
 			bluetooth_chooser_start_discovery (self);
 			priv->disco_rq = FALSE;
 		}
-
-		g_signal_connect (priv->adapter_model, "row-changed",
-				  G_CALLBACK (device_model_row_changed), self);
 	}
 }
 
@@ -476,12 +560,21 @@ create_treeview (BluetoothChooser *self)
 	gtk_tree_view_column_set_attributes (column, renderer,
 					     "text", BLUETOOTH_COLUMN_ALIAS, NULL);
 
+	/* The connected icon */
+	priv->connected_cell = gtk_cell_renderer_pixbuf_new ();
+	gtk_tree_view_column_pack_start (column, priv->connected_cell, FALSE);
+
+	gtk_tree_view_column_set_cell_data_func (column, priv->connected_cell,
+						 connected_to_icon, NULL, NULL);
+	g_object_set (G_OBJECT (priv->connected_cell), "visible", priv->show_connected, NULL);
+
 	/* The bonded icon */
 	priv->bonded_cell = gtk_cell_renderer_pixbuf_new ();
 	gtk_tree_view_column_pack_end (column, priv->bonded_cell, FALSE);
 
 	gtk_tree_view_column_set_cell_data_func (column, priv->bonded_cell,
 						 bonded_to_icon, NULL, NULL);
+	g_object_set (G_OBJECT (priv->bonded_cell), "visible", priv->show_paired, NULL);
 
 	gtk_tree_view_append_column (GTK_TREE_VIEW(tree), column);
 
@@ -505,10 +598,9 @@ create_treeview (BluetoothChooser *self)
 		gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER (priv->filter),
 							filter_func, self, NULL);
 		gtk_tree_view_set_model (GTK_TREE_VIEW(tree), priv->filter);
-		g_object_unref (priv->filter);
-
-		g_signal_connect (priv->adapter_model, "row-changed",
+		g_signal_connect (priv->filter, "row-changed",
 				  G_CALLBACK (device_model_row_changed), self);
+		g_object_unref (priv->filter);
 	} else {
 		gtk_widget_set_sensitive (GTK_WIDGET (tree), FALSE);
 		gtk_widget_set_sensitive (GTK_WIDGET (priv->search_button), FALSE);
@@ -533,6 +625,9 @@ bluetooth_chooser_init(BluetoothChooser *self)
 	GtkWidget *hbox;
 	GtkWidget *scrolled_window;
 	GtkWidget *table;
+	GtkCellRenderer *renderer;
+
+	gtk_widget_push_composite_child ();
 
 	priv->show_paired = FALSE;
 	priv->show_search = FALSE;
@@ -549,7 +644,7 @@ bluetooth_chooser_init(BluetoothChooser *self)
 
 	/* The top level label */
 	priv->label = gtk_label_new ("");
-	gtk_widget_show (priv->label);
+	gtk_widget_set_no_show_all (priv->label, TRUE);
 	gtk_box_pack_start (GTK_BOX (vbox), priv->label, FALSE, TRUE, 0);
 	gtk_label_set_use_markup (GTK_LABEL (priv->label), TRUE);
 	gtk_misc_set_alignment (GTK_MISC (priv->label), 0, 0.5);
@@ -595,6 +690,7 @@ bluetooth_chooser_init(BluetoothChooser *self)
 	gtk_widget_show (vbox);
 	gtk_box_pack_start (GTK_BOX (self), vbox, FALSE, TRUE, 0);
 	priv->filters_vbox = vbox;
+	gtk_widget_set_no_show_all (priv->filters_vbox, TRUE);
 
 	/* The filters */
 	str = g_strdup_printf ("<b>%s</b>", _("Show Only Bluetooth Devices With..."));
@@ -655,21 +751,44 @@ bluetooth_chooser_init(BluetoothChooser *self)
 	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
 	priv->device_type_label = label;
 
-	priv->device_type = gtk_combo_box_new_text ();
+	priv->device_type_filter_model = GTK_TREE_MODEL (gtk_list_store_new (DEVICE_TYPE_FILTER_NUM_COLS,
+									     G_TYPE_STRING, G_TYPE_INT));
+	priv->device_type = gtk_combo_box_new_with_model (priv->device_type_filter_model);
+	renderer = gtk_cell_renderer_text_new ();
+	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (priv->device_type), renderer, TRUE);
+	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (priv->device_type), renderer, "text", DEVICE_TYPE_FILTER_COL_NAME);
+
 	gtk_widget_set_no_show_all (priv->device_type, TRUE);
 	gtk_widget_show (priv->device_type);
 	gtk_table_attach (GTK_TABLE (table), priv->device_type, 1, 2, 0, 1,
 			  (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
 			  (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);
 	gtk_widget_set_tooltip_text (priv->device_type, _("Select the device type to filter above list"));
-	/* The types match the types used in client.h */
-	for (i = 0; i < _BLUETOOTH_TYPE_NUM_TYPES; i++) {
-		gtk_combo_box_append_text (GTK_COMBO_BOX(priv->device_type),
-					   _(bluetooth_type_to_string (1 << i)));
+	gtk_list_store_insert_with_values (GTK_LIST_STORE (priv->device_type_filter_model), NULL, G_MAXUINT32,
+					   DEVICE_TYPE_FILTER_COL_NAME, _(bluetooth_type_to_string (BLUETOOTH_TYPE_ANY)),
+					   DEVICE_TYPE_FILTER_COL_MASK, BLUETOOTH_TYPE_ANY,
+					   -1);
+	gtk_list_store_insert_with_values (GTK_LIST_STORE (priv->device_type_filter_model), NULL, G_MAXUINT32,
+					   DEVICE_TYPE_FILTER_COL_NAME, _("Input devices (mice, keyboards, ...)"),
+					   DEVICE_TYPE_FILTER_COL_MASK, BLUETOOTH_TYPE_INPUT,
+					   -1);
+	gtk_list_store_insert_with_values (GTK_LIST_STORE (priv->device_type_filter_model), NULL, G_MAXUINT32,
+					   DEVICE_TYPE_FILTER_COL_NAME, _("Headphones, headsets and other audio devices"),
+					   DEVICE_TYPE_FILTER_COL_MASK, BLUETOOTH_TYPE_AUDIO,
+					   -1);
+	/* The types match the types used in bluetooth-client.h */
+	for (i = 1; i < _BLUETOOTH_TYPE_NUM_TYPES; i++) {
+		int mask = 1 << i;
+		if (mask & BLUETOOTH_TYPE_INPUT || mask & BLUETOOTH_TYPE_AUDIO)
+			continue;
+		gtk_list_store_insert_with_values (GTK_LIST_STORE (priv->device_type_filter_model), NULL, G_MAXUINT32,
+						   DEVICE_TYPE_FILTER_COL_NAME, _(bluetooth_type_to_string (mask)),
+						   DEVICE_TYPE_FILTER_COL_MASK, mask,
+						   -1);
 	}
 	g_signal_connect (G_OBJECT (priv->device_type), "changed",
 			  G_CALLBACK(filter_type_changed_cb), self);
-	gtk_combo_box_set_active (GTK_COMBO_BOX(priv->device_type), int_log2(priv->device_type_filter));
+	set_combobox_from_filter (self);
 	if (priv->show_device_type) {
 		gtk_widget_show (priv->device_type_label);
 		gtk_widget_show (priv->device_type);
@@ -681,6 +800,8 @@ bluetooth_chooser_init(BluetoothChooser *self)
 
 	priv->default_adapter_changed_id = g_signal_connect (priv->client, "notify::default-adapter",
 							     G_CALLBACK (default_adapter_changed), self);
+
+	gtk_widget_pop_composite_child ();
 }
 
 static void
@@ -715,7 +836,10 @@ enum {
 	PROP_DEVICE_SELECTED_ICON,
 	PROP_DEVICE_SELECTED_NAME,
 	PROP_DEVICE_SELECTED_TYPE,
+	PROP_DEVICE_SELECTED_IS_CONNECTED,
+	PROP_DEVICE_SELECTED_PROXY,
 	PROP_SHOW_PAIRING,
+	PROP_SHOW_CONNECTED,
 	PROP_SHOW_SEARCH,
 	PROP_SHOW_DEVICE_TYPE,
 	PROP_SHOW_DEVICE_CATEGORY,
@@ -738,6 +862,11 @@ bluetooth_chooser_set_property (GObject *object, guint prop_id,
 		if (priv->bonded_cell != NULL)
 			g_object_set (G_OBJECT (priv->bonded_cell), "visible", priv->show_paired, NULL);
 		break;
+	case PROP_SHOW_CONNECTED:
+		priv->show_connected = g_value_get_boolean (value);
+		if (priv->connected_cell != NULL)
+			g_object_set (G_OBJECT (priv->connected_cell), "visible", priv->show_connected, NULL);
+		break;
 	case PROP_SHOW_SEARCH:
 		priv->show_search = g_value_get_boolean (value);
 		g_object_set (G_OBJECT (priv->search_button), "visible", priv->show_search, NULL);
@@ -762,7 +891,7 @@ bluetooth_chooser_set_property (GObject *object, guint prop_id,
 		break;
 	case PROP_DEVICE_TYPE_FILTER:
 		priv->device_type_filter = g_value_get_int (value);
-		gtk_combo_box_set_active (GTK_COMBO_BOX(priv->device_type), int_log2(priv->device_type_filter));
+		set_combobox_from_filter (BLUETOOTH_CHOOSER (object));
 		break;
 	case PROP_DEVICE_CATEGORY_FILTER:
 		priv->device_category_filter = g_value_get_int (value);
@@ -794,9 +923,24 @@ bluetooth_chooser_get_property (GObject *object, guint prop_id,
 	case PROP_DEVICE_SELECTED_TYPE:
 		g_value_set_uint (value, bluetooth_chooser_get_selected_device_type (self));
 		break;
+	case PROP_DEVICE_SELECTED_IS_CONNECTED:
+		g_value_set_boolean (value, bluetooth_chooser_get_selected_device_is_connected (self));
+		break;
+	case PROP_DEVICE_SELECTED_PROXY: {
+		GtkTreeIter iter;
+		GObject *proxy = NULL;
+
+		if (gtk_tree_selection_get_selected (priv->selection, NULL, &iter) != FALSE)
+			gtk_tree_model_get (priv->filter, &iter, BLUETOOTH_COLUMN_PROXY, &proxy, -1);
+		g_value_take_object (value, proxy);
+		break;
+	}
 	case PROP_SHOW_PAIRING:
 		g_value_set_boolean (value, priv->show_paired);
 		break;
+	case PROP_SHOW_CONNECTED:
+		g_value_set_boolean (value, priv->show_connected);
+		break;
 	case PROP_SHOW_SEARCH:
 		g_value_set_boolean (value, priv->show_search);
 		break;
@@ -821,6 +965,11 @@ bluetooth_chooser_get_property (GObject *object, guint prop_id,
 static void
 bluetooth_chooser_class_init (BluetoothChooserClass *klass)
 {
+	/* Use to calculate the maximum value for the 
+	 * device-type-filter value */
+	guint i;
+	int max_filter_val;
+
 	g_type_class_add_private(klass, sizeof(BluetoothChooserPrivate));
 
 	G_OBJECT_CLASS(klass)->finalize = bluetooth_chooser_finalize;
@@ -886,6 +1035,18 @@ bluetooth_chooser_class_init (BluetoothChooserClass *klass)
 	g_object_class_install_property (G_OBJECT_CLASS(klass),
 					 PROP_DEVICE_SELECTED_TYPE, g_param_spec_uint ("device-selected-type", NULL, NULL,
 										       1, 1 << (_BLUETOOTH_TYPE_NUM_TYPES - 1), 1, G_PARAM_READABLE));
+	/**
+	 * BluetoothChooser:device-selected-is-connected:
+	 *
+	 * whether the selected device is conncted to this computer, will be %FALSE if no devices are selected
+	 **/
+	g_object_class_install_property (G_OBJECT_CLASS(klass),
+					 PROP_DEVICE_SELECTED_IS_CONNECTED, g_param_spec_boolean ("device-selected-is-connected", NULL, NULL,
+												  FALSE, G_PARAM_READABLE));
+	/* Left blank intentionally */
+	g_object_class_install_property (G_OBJECT_CLASS(klass),
+					 PROP_DEVICE_SELECTED_PROXY, g_param_spec_object ("device-selected-proxy", NULL, NULL,
+												  G_TYPE_OBJECT, G_PARAM_READABLE));
 	/**
 	 * BluetoothChooser:show-pairing:
 	 *
@@ -894,6 +1055,14 @@ bluetooth_chooser_class_init (BluetoothChooserClass *klass)
 	g_object_class_install_property (G_OBJECT_CLASS(klass),
 					 PROP_SHOW_PAIRING, g_param_spec_boolean ("show-pairing",
 										  NULL, NULL, FALSE, G_PARAM_READWRITE));
+	/**
+	 * BluetoothChooser:show-connected:
+	 *
+	 * Whether to show the connected column in the tree.
+	 **/
+	g_object_class_install_property (G_OBJECT_CLASS(klass),
+					 PROP_SHOW_CONNECTED, g_param_spec_boolean ("show-connected",
+										    NULL, NULL, FALSE, G_PARAM_READWRITE));
 	/**
 	 * BluetoothChooser:show-search:
 	 *
@@ -924,9 +1093,11 @@ bluetooth_chooser_class_init (BluetoothChooserClass *klass)
 	 *
 	 * FIXME
 	 **/
+	for (i = 0, max_filter_val = 0 ; i < _BLUETOOTH_TYPE_NUM_TYPES; i++)
+		max_filter_val += 1 << i;
 	g_object_class_install_property (G_OBJECT_CLASS(klass),
 					 PROP_DEVICE_TYPE_FILTER, g_param_spec_int ("device-type-filter", NULL, NULL,
-										    1, 1 << (_BLUETOOTH_TYPE_NUM_TYPES - 1), 1, G_PARAM_READWRITE));
+										    1, max_filter_val, 1, G_PARAM_READWRITE));
 	/**
 	 * BluetoothChooser:device-category-filter:
 	 *
